/* gs.c: game sorter */

/* Revised: 1993.04.09 */

/*

This program sorts a file of PGN (Portable Game Notation) games from
the standard input and places the result on the standard output.  Note
that this program is fairly strict about the input; it is intended for
use with files produced by the SAN Kit or by other programs that
generate Seven Tag Generic Format PGN.

The source file for this program usually resides in the SAN_DOC
directory in the SAN Kit package.

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef NULL
#undef NULL
#endif
#define NULL ((void *) 0)

#define nonstatic

typedef short int siT;
typedef long int liT;
typedef char *charptrT;

#define bL 256

#define giL 7

#define gi_event  0
#define gi_site   1
#define gi_date   2
#define gi_round  3
#define gi_white  4
#define gi_black  5
#define gi_result 6

typedef struct gS
	{
	struct gS *g_lptr;
	struct gS *g_rptr;
	charptrT g_tagv[giL];
	charptrT g_score;
	} gT, *gptrT;

static charptrT progname;
static charptrT gisv[giL];

static liT line_n, game_n;
static gptrT grptr;

static char bv[bL];

/*--> Fatal: report fatal error and exit */
static
void
Fatal(charptrT s)
{
fprintf(stderr, "%s: fatal error: %s\n", progname, s);
exit(1);

return;
}

/*--> FatalLoc: report fatal error with location and exit */
static
void
FatalLoc(charptrT s)
{
fprintf(stderr, "%s: fatal error: %s (line: %ld   game: %ld)\n",
	progname, s, line_n, game_n);
exit(1);

return;
}

/*--> ReadLine: read a line from standard input */
static
charptrT
ReadLine(void)
{
charptrT ptr;

ptr = fgets(bv, bL, stdin);
if (ptr != NULL)
	line_n++;

return (ptr);
}

/*--> GrabStr: allocate a string */
static
charptrT
GrabStr(charptrT s)
{
charptrT p;
liT length;

length = strlen(s);
p = malloc(length + 1);
if (p == NULL)
	FatalLoc("out of memory");

strcpy(p, s);

return (p);
}

/*--> ConcatStr: special concatenation */
static
charptrT
ConcatStr(charptrT sptr0, charptrT sptr1)
{
charptrT sptr;
liT length;

length = strlen(sptr0) + strlen(sptr1);
sptr = malloc(length + 1);
if (sptr == NULL)
	Fatal("out of memory");

strcpy(sptr, sptr0);
strcat(sptr, sptr1);
free(sptr0);

return (sptr);
}

/*--> NewGame: allocate a new game record */
static
gptrT
NewGame(void)
{
gptrT gptr;
siT gi;

gptr = (gptrT) malloc(sizeof(gT));
if (gptr == NULL)
	FatalLoc("out of memory");

gptr->g_lptr = gptr->g_rptr = NULL;
for (gi = 0; gi < giL; gi++)
	gptr->g_tagv[gi] = NULL;
gptr->g_score = NULL;

return (gptr);
}

/*--> ReleaseGame: deallocate a game record */
static
void
ReleaseGame(gptrT gptr)
{
siT gi;

if (gptr != NULL)
	{
	ReleaseGame(gptr->g_lptr);
	ReleaseGame(gptr->g_rptr);
	if (gptr->g_score != NULL)
	  free(gptr->g_score);
	for (gi = 0; gi < giL; gi++)
		if (gptr->g_tagv[gi] != NULL)
			free(gptr->g_tagv[gi]);
	free(gptr);
	};

return;
}

/*--> CompGame: compare two games */
static
siT
CompGame(gptrT gptr0, gptrT gptr1)
{
siT result, result_save;
siT c0, c1;
liT r0, r1;
liT l0, l1;

result = 0;

if (result == 0)
	result = strcmp(gptr0->g_tagv[gi_date], gptr1->g_tagv[gi_date]);

if (result == 0)
	result = strcmp(gptr0->g_tagv[gi_event], gptr1->g_tagv[gi_event]);

if (result == 0)
	{
	result = strcmp(gptr0->g_tagv[gi_round], gptr1->g_tagv[gi_round]);
	if (result != 0)
		{
		result_save = result;
		l0 = strlen(gptr0->g_tagv[gi_round]);
		l1 = strlen(gptr1->g_tagv[gi_round]);
		if ((l0 < 1) || (l1 < 1))
			result = 0;
		else
			{
			c0 = sscanf(gptr0->g_tagv[gi_round], "%ld", &r0);
			c1 = sscanf(gptr1->g_tagv[gi_round], "%ld", &r1);

			if (c0 == 1)
				if (c1 == 1)
					{
					if (r0 == r1)
						result = 0;
					else
						{
						if (r0 < r1)
							result = -1;
						else
							result = 1;
						};
					}
				else
					result = 1;
			else
				{
				if (c1 == 1)
					result = -1;
				else
					result = result_save;
				};
			};
		};
	};

if (result == 0)
	result = strcmp(gptr0->g_tagv[gi_site], gptr1->g_tagv[gi_site]);

if (result == 0)
	result = strcmp(gptr0->g_tagv[gi_white], gptr1->g_tagv[gi_white]);

if (result == 0)
	result = strcmp(gptr0->g_tagv[gi_black], gptr1->g_tagv[gi_black]);

if (result == 0)
	result = strcmp(gptr0->g_tagv[gi_result], gptr1->g_tagv[gi_result]);

if (result == 0)
	result = strcmp(gptr0->g_score, gptr1->g_score);

return (result);
}

/*--> InsertGame: insert a game into the tree */
static
void
InsertGame(gptrT gptr)
{
siT flag, result;
gptrT xptr;

if (grptr == NULL)
	grptr = gptr;
else
	{
	flag = 0;
	xptr = grptr;
	while (!flag)
		{
		result = CompGame(xptr, gptr);
		if (result != 0)
			{
			if (result > 0)
				{
				if (xptr->g_lptr != NULL)
					xptr = xptr->g_lptr;
				else
					{
					xptr->g_lptr = gptr;
					flag = 1;
					};
				}
			else
				{
				if (xptr->g_rptr != NULL)
					xptr = xptr->g_rptr;
				else
					{
					xptr->g_rptr = gptr;
					flag = 1;
					};
				};
			}
		else
			{
			fprintf(stderr,
				"%s: duplicate game (line: %ld / game: %ld) deleted\n",
				progname, line_n, game_n);
			ReleaseGame(gptr);
			flag = 1;
			};
		};
	};

return;
}

/*--> ReadGame: read a game and return record */
static
gptrT
ReadGame(void)
{
gptrT gptr;
siT flag;
siT gi;
liT length0, length1, length2;

flag = 1;
gptr = NewGame();
gptr->g_lptr = gptr->g_rptr = NULL;
gptr->g_score = NULL;
for (gi = 0; gi < giL; gi++)
	gptr->g_tagv[gi] = NULL;

game_n++;

gi = 0;
while (flag && (gi < giL))
	{
	if (ReadLine() == NULL)
		if (gi == 0)
			flag = 0;
		else
			FatalLoc("unexpected EOF");

	if (flag)
		{
		length0 = strlen(bv);
		length1 = strlen(gisv[gi]);
		length2 = length0 - (length1 + 6);
		if (length2 < 0)
			FatalLoc("tag fault");
		};

	if (flag)
		{
		gptr->g_tagv[gi] = malloc(length2 + 1);
		if (gptr->g_tagv[gi] == NULL)
			FatalLoc("out of memory");
		strncpy(gptr->g_tagv[gi], &bv[length1 + 3], length2);
		*(gptr->g_tagv[gi] + length2) = '\0';
		gi++;
		};
	};

if (flag)
	{
	if (ReadLine() == NULL)
		FatalLoc("unexpected EOF");

	if (strlen(bv) != 1)
		FatalLoc("expected blank line");
	};

if (flag)
	{
	gptr->g_score = GrabStr("");
	do
		{
		if (ReadLine() == NULL)
			FatalLoc("unexpected EOF");

		gptr->g_score = ConcatStr(gptr->g_score, bv);
		}
	while (flag && (strlen(bv) > 1));
	};

if (!flag)
	{
	ReleaseGame(gptr);
	gptr = NULL;
	};

if (gptr == NULL)
	game_n--;

return (gptr);
}

/*--> WriteGame: write a game from a record */
static
void
WriteGame(gptrT gptr)
{
siT gi;

for (gi = 0; gi < giL; gi++)
	printf("[%s \"%s\"]\n", gisv[gi], gptr->g_tagv[gi]);
printf("\n");
printf("%s", gptr->g_score);

return;
}

/*--> WriteNode: write the game tree from a node */
static
void
WriteNode(gptrT gptr)
{
if (gptr != NULL)
	{
	WriteNode(gptr->g_lptr);
	WriteGame(gptr);
	WriteNode(gptr->g_rptr);
	};

return;
}

/*--> main: main program for game sorter */
nonstatic
int
main(int argc, charptrT argv[])
{
gptrT gptr;

progname = argv[0];

if (argc != 1)
	Fatal("filter mode only");

/* initialize */

gisv[gi_event] = "Event";
gisv[gi_site] = "Site";
gisv[gi_date] = "Date";
gisv[gi_round] = "Round";
gisv[gi_white] = "White";
gisv[gi_black] = "Black";
gisv[gi_result] = "Result";

line_n = game_n = 0;
grptr = NULL;

/* read games */

while ((gptr = ReadGame()) != NULL)
	InsertGame(gptr);

/* write games */

WriteNode(grptr);

/* deallocate */

ReleaseGame(grptr);
grptr = NULL;

/* report */

fprintf(stderr, "%ld games   %ld lines\n", game_n, line_n);

return (0);
}

/* gs.c: EOF */
